"""
This module contains project meta-data.
"""

__author__ = "Caleb P. Burns"
__copyright__ = "Copyright © 2012-2023 by Caleb P. Burns"
__credits__ = [
	"khomyakov42 <https://github.com/khomyakov42>",
	"pedermoller <https://github.com/pedermoller>",
	"ryokash <https://github.com/ryokash>",
	"zenwalker <https://github.com/zenwalker>",
	"kesmit13 <https://github.com/kesmit13>",
	"Cnoor0171 <https://github.com/Cnoor0171>",
	"sergey-shambir <https://github.com/sergey-shambir>",
]
__license__ = "MIT License"
__version__ = "5.1.0"
